package common;
import java.io.InputStream;
import java.math.BigInteger;
import java.util.Random;

/**
 * Provides encryption and decryption operations.
 * <p>
 * The used encryption scheme is
 * <p>
 * \f$y=E(x)=b\cdot{}(r\cdot{}q+x)\bmod{p}\f$
 * <p>
 * The decryption is done by
 * <p>
 * \f$x=D(y)=b^{-1}\cdot{}y\bmod{p}\bmod{q}\f$
 * <p>
 * If an encrypted value is known to contain <code>b</code> of a degree <code>k</code> greater than 1,
 * the decryption is instead done by
 * \f$x=D(y,k)=b^{-k}\cdot{}y\bmod{p}\bmod{q}\f$
 * <p>
 * Before using, the Cipher object needs to be initialized with either a parameters file (see {@link #initialize(String)})
 * or a stream (see {@link #initialize(InputStream)}) containing the parameters.
 * 
 * @author vohuudtr
 *
 */
public class Cipher {
	private static BigInteger b_inverse = new BigInteger("0");
	private static Random seed_r = new Random();

	/**
	 * Initializes the cipher with parameters in a given file.
	 * @param paramfile parameters filename.
	 */
	public static void initialize(String paramfile) {
		Parameters.readConfig(paramfile);
		initialize();
	}
	
	/**
	 * Initializes the cipher with parameters in a stream.
	 * @param is stream containing the parameters.
	 */
	public static void initialize(InputStream is) {
		Parameters.readConfig(is);
		initialize();
	}
	
	/**
	 * Creates a random object and the inverse of <code>b</code>.
	 */
	private static void initialize() {
		seed_r = new Random(System.currentTimeMillis());
		b_inverse = Parameters.getB().modInverse(Parameters.getP());
	}

	/**
	 * Encrypts an integer <code>x</code> to <code>E(x)</code>.
	 * 
	 * @param x integer to be encrypted.
	 * @return encrypted integer.
	 */
	public static BigInteger encrypt(BigInteger x) {
		BigInteger r;

		r = new BigInteger(Parameters.getS2(), seed_r);
		return r.multiply(Parameters.getQ()).add(x).multiply(Parameters.getB()).mod(Parameters.getP());
	}

	/**
	 * Decrypts an encrypted value to the original integer value.
	 * 
	 * @param c encrypted value.
	 * @return plain-text integer.
	 */
	public static BigInteger decrypt(BigInteger c) {
		return c.multiply(b_inverse).mod(Parameters.getP()).mod(Parameters.getQ());
	}
	
	/**
	 * Decrypts an encrypted value of a specified degree of <code>b</code> to the original integer value.
	 * @param c encrypted value.
	 * @param degree known degree of <code>b</code> in the encrypted value.
	 * @return plain-text integer.
	 */
	public static BigInteger decrypt(BigInteger c, int degree) {
		return c.multiply(b_inverse.pow(degree)).mod(Parameters.getP()).mod(Parameters.getQ());
	}


}
