package common;

import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigInteger;

import mapred.BigIntegerWritable;

/**
 * Computes the user's request as a set of encrypted coefficients corresponding
 * to the queried indicator polynomial.
 * <p>
 * The current implementation supports computing the query only for a specified value
 * over given fields. General boolean expressions are not supported, thus should be
 * done manually.
 * <p>
 * Usage:
 * <p>
 * <code>java %common.ClientRequest &lt;mapred&gt; &lt;key&gt; &lt;paramfile&gt; &lt;requestfile&gt;</code>
 * <dl>
 * <dt>mapred</dt>
 * <dd>
 * <dl>
 * <dt>epic</dt><dd>Make request for {@link mapred.MapRedEpic MapRedEpic}.</dd>
 * <dt>plain</dt><dd>Make request for {@link mapred.MapRedPlainCountOne MapRedPlainCountOne}.</dd>
 * <dt>notsendcoeff</dt><dd>Make request for {@link mapred.MapRedNotSendCoeff MapRedNotSendCoeff}.</dd>
 * </dl>
 * </dd>
 * <dt>key</dt><dd>Value to be counted.</dd>
 * <dt>paramfile</dt><dd>Path to the parameters file (see {@link Parameters}).</dd>
 * <dt>requestfile</dt><dd>Path to the request to be created.</dd>
 * </dl>
 * @author vohuudtr
 * 
 */
public class ClientRequest {

	private static DataOutput getOutput(String filename) {
		FileOutputStream fos;
		try {
			fos = new FileOutputStream(filename);
			DataOutputStream dos = new DataOutputStream(fos);
			return dos;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}

		return null;
	}

	private static int[] domain(int size) {
		int[] x = new int[size];
		for (int k = 0; k < x.length; k++)
			x[k] = k;
		return x;
	}

	private static void computeCoeffs(Polynomial[] f,
			MultiArray<BigIntegerWritable> coeffs, MultiFieldKey runningKey,
			BigInteger runningCoeff, int runningField) {
		for (int v = 0; v < Parameters.getSizeD(runningField); v++) {
			BigInteger c = f[runningField].getCoefficient(v);
			if (c.signum() != 0) {
				runningKey.element[runningField] = v;
				if (runningField == Parameters.getNumFields() - 1) {
					coeffs.put(runningKey, new BigIntegerWritable(runningCoeff
							.multiply(c).mod(Parameters.getQ())));
				} else {
					computeCoeffs(f, coeffs, runningKey,
							runningCoeff.multiply(c).mod(Parameters.getQ()),
							runningField + 1);
				}
			}
		}
	}

	/**
	 * Makes request for {@link mapred.MapRedEpic MapRedEpic}.
	 * @param key value to be counted.
	 * @param out output stream containing the created request.
	 */
	public static void requestEpic(MultiFieldKey key, DataOutput out) {
		// create Lagrange polynomial for each field
		Polynomial[] f = new Polynomial[key.element.length];
		for (int field = 0; field < key.element.length; field++) {
			f[field] = Polynomial.createLagrange(
					domain(Parameters.getSizeD(field)), key.element[field]);
//			System.out.println("f " + f[field]);
		}

		// compute combined polynomial's coefficients
		MultiArray<BigIntegerWritable> coeffs = MultiArray
				.createBigIntegerWritableArray(Parameters.getTotalSizeD());
		computeCoeffs(f, coeffs, new MultiFieldKey(Parameters.getNumFields()),
				new BigInteger("1"), 0);

		int maxweight = MultiArray.getMultiFieldKey(coeffs.getSize() - 1)
				.weight();
		// encrypt the coefficients before sending to the cloud
		for (int i = 0; i < coeffs.getSize(); i++) {
			BigIntegerWritable c = coeffs.get(i);
			if (c == null)
				c = new BigIntegerWritable(0);
			BigInteger enc_c = Cipher.encrypt(new BigInteger(c.toString()));
			BigInteger one = Cipher.encrypt(new BigInteger("1"));
//			System.out.println(maxweight - MultiArray.getMultiFieldKey(i).weight());
			one = one.pow(maxweight - MultiArray.getMultiFieldKey(i).weight());
			enc_c = enc_c.multiply(one).mod(Parameters.getP());
			coeffs.put(i, new BigIntegerWritable(enc_c));
		}

		// write to the request file
		try {
			MultiArray.write(coeffs, out);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Makes request for {@link mapred.MapRedPlainCountOne MapRedPlainCountOne}.
	 * @param key value to be counted.
	 * @param out output stream containing the created request.
	 */
	public static void requestPlain(MultiFieldKey key, DataOutput out) {
		// write to the request file
		try {
			key.write(out);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Makes request for {@link mapred.MapRedNotSendCoeff MapRedNotSendCoeff}.
	 * This is actually a <b>fake</b> method as we do not send any coefficients
	 * when using {@link mapred.MapRedNotSendCoeff MapRedNotSendCoeff}. Processing
	 * of the query in this case is actually done by
	 * {@link LocalCount LocalCount}, which handles the request and answer locally.
	 * So the arguments provided to this method can be <code>null</code>.
	 * @param key value to be counted, can be <code>null</code>.
	 * @param out output stream containing the created request. The output stream can be <code>null</code>.
	 */
	public static void requestNotSendCoeff(MultiFieldKey key, DataOutput out) {
		// do nothing because the request does not contain coefficients
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String countingType = null, countedKey = null, paramfile = null, requestfile = null;

		if (args.length > 0) {
			countingType = args[0];
			countedKey = args[1];
			paramfile = args[2];
			requestfile = args[3];
		}

		if (args.length == 0
				|| !(countingType.equals("epic")
						|| countingType.equals("plain") || countingType
							.equals("notsendcoeff"))) {
			System.out.println("Client makes counting request to the cloud.");
			System.out
					.println("Usage: java "
							+ ClientRequest.class.getName()
							+ " <epic|plain|notsendcoeff> <key> <paramfile> <requestfile>");
			System.exit(0);
		}

		// parse requested key
		String[] ind = countedKey.split(",");
		MultiFieldKey key = new MultiFieldKey(ind.length);
		for (int i = 0; i < key.element.length; i++)
			key.element[i] = Integer.parseInt(ind[i]);

		// create request
		if (countingType.equals("epic")) {
			Cipher.initialize(paramfile);
			requestEpic(key, getOutput(requestfile));
		} else if (countingType.equals("notsendcoeff")) {
			requestNotSendCoeff(key, getOutput(requestfile));
		} else if (countingType.equals("plain")) {
			requestPlain(key, getOutput(requestfile));
		}
	}

}
