package common;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.math.BigInteger;

/**
 * Provides a tool to obtain the plain-text count value from the answer received from the cloud.
 * The answer is simply decrypted using the known degree of <code>b</code>.
 * <p>
 * Usage:
 * <p>
 * <code>java %common.GetAnswer &lt;mapred&gt; &lt;paramfile&gt; &lt;answerfile&gt;</code>
 * <dl>
 * <dt>mapred</dt>
 * <dd>
 * <dl>
 * <dt>epic</dt><dd>Gets answer received from {@link mapred.MapRedEpic MapRedEpic}.</dd>
 * <dt>plain</dt><dd>Gets answer received from {@link mapred.MapRedPlainCountOne MapRedPlainCountOne}.</dd>
 * </dl>
 * </dd>
 * <dt>paramfile</dt><dd>Path to the parameters file (see {@link Parameters}).</dd>
 * <dt>answerfile</dt><dd>Path to the answer file.</dd>
 * </dl>
 * @author vohuudtr
 * 
 */
public class GetAnswer {

	@SuppressWarnings("deprecation")
	private static String getAnswer(String filename) {
		FileInputStream fis;
		try {
			fis = new FileInputStream(filename);
			DataInputStream dis = new DataInputStream(fis);
			String answer = dis.readLine();
			dis.close();
			fis.close();
			return answer;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return null;
	}

	/**
	 * Decrypts the encrypted answer to the plain-text count value.
	 * The plain-text value is printed to the standard output.
	 * @param answer encrypted answer as a big-integer in Java String format.
	 */
	public static void parseAnswerEpic(String answer) {
		BigInteger c = new BigInteger(answer);
		int degree = 0;

		for (int field = 0; field < Parameters.getNumFields(); field++)
			degree += Parameters.getSizeD(field) - 1;
		degree++;

		BigInteger count = Cipher.decrypt(c, degree);
		System.out.println(count);
	}

	/**
	 * In plain-text MapReduce cases, the received answer is also the count value.
	 * The answer is, therefore, printed directly to the standard output.
	 * @param answer plain-text answer as a big-integer in Java String format.
	 */
	public static void parseAnswerPlain(String answer) {
		System.out.println(answer);
	}

	private static void usage() {
		System.out.println("Parse answer from the cloud.");
		System.out.println("Usage: java " + GetAnswer.class.getName()
				+ " <epic|plain> <paramfile> <answerfile>");
		System.exit(0);
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String countingType = null, paramfile = null, answerfile = null;

		if (args.length != 3)
			usage();

		countingType = args[0];
		paramfile = args[1];
		answerfile = args[2];

		if (countingType.equals("epic")) {
			Cipher.initialize(paramfile);
			parseAnswerEpic(getAnswer(answerfile));
		} else if (countingType.equals("plain")) {
			parseAnswerPlain(getAnswer(answerfile));
		}
	}

}
