package common;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.math.BigInteger;

/**
 * Provides tools for handling the parameters used for the 
 * {@link Cipher encryption scheme}.
 * <p>
 * The following parameters are controlled:
 * <dl>
 * <dt>Secret key</dt>
 * <dd>
 * <dl>
 * <dt>p</dt><dd>secret prime <code>p</code> in the encryption scheme.</dd>
 * <dt>b</dt><dd>secret random <code>b</code> in the encryption scheme.</dd>
 * </dl>
 * </dd>
 * <dt>Public parameters</dt>
 * <dd>
 * <dl>
 * <dt>q</dt><dd>prime <code>q</code> in the encryption scheme.</dd>
 * <dt>s1</dt><dd>security parameter <code>s1</code> in the encryption scheme.</dd>
 * <dt>s2</dt><dd>maximum size in bits of the random <code>r</code> in the encryption scheme.</dd>
 * <dt>n</dt><dd>upperbound of the number of records in the data set.</dd>
 * <dt>numFields</dt><dd>number of countable fields in each record.<br>
 * <em>Note: </em><b>numFields</b> is the actual number of countable fields in a record, while
 * {@link Record#getNumFields()} returns the number of fields contained in a record, which includes
 * both encrypted and plain-text fields, so {@link Record#getNumFields()} is equal to <code>2*numFields</code>
 * (see {@link Record}).</dd>
 * <dt>sizeD[]</dt><dd>Domain size in bits of the countable fields.</dd>
 * </dl>
 * </dd>
 * </dl>
 * @author vohuudtr
 *
 */
public class Parameters {
	private static BigInteger p = new BigInteger("0"); // (secret) prime p
	private static BigInteger b = new BigInteger("0"); // (secret) random b
	private static BigInteger q = new BigInteger("0"); // (public) prime q
	private static BigInteger n = new BigInteger("0"); // (public) number of records in dataset
	private static int s1; // (public) security parameter s1
	private static int s2; // (public) security parameter s2, size of random r
	private static int numFields; // (public) number of countable fields
	private static int[] size_D; // (public) size of countable fields' domain, i.e. |D|
	private static int maxSize_D;

	private static long t = System.currentTimeMillis();
	private static long delta_t;

	/* Getters and setters */

	/**
	 * Returns secret prime <code>p</code>.
	 * @return secret prime <code>p</code>.
	 */
	public static BigInteger getP() {
		return p;
	}

	/**
	 * Sets new secret prime <code>p</code>.
	 * @param newp new secret prime <code>p</code>.
	 */
	public static void setP(BigInteger newp) {
		p = newp;
	}

	/**
	 * Returns secret <code>b</code>. 
	 * @return secret <code>b</code>.
	 */
	public static BigInteger getB() {
		return b;
	}

	/**
	 * Sets the new value for <code>b</code>.
	 * @param newb new value for <code>b</code>.
	 */
	public static void setB(BigInteger newb) {
		b = newb;
	}
	
	/**
	 * Returns prime <code>q</code>.
	 * @return prime <code>q</code>.
	 */
	public static BigInteger getQ() {
		return q;
	}
	
	/**
	 * Sets new prime <code>q</code>.
	 * @param newq new prime <code>q</code>.
	 */
	public static void setQ(BigInteger newq) {
		q = newq;
	}

	/**
	 * Returns upperbound of the number of records.
	 * @return upperbound of the number of records.
	 */
	public static BigInteger getN() {
		return n;
	}

	/**
	 * Returns security parameter <code>s1</code>.
	 * @return security parameter <code>s1</code>.
	 */
	public static int getS1() {
		return s1;
	}

	/**
	 * Returns security parameter <code>s2</code>.
	 * @return security parameter <code>s2</code>.
	 */
	public static int getS2() {
		return s2;
	}

	/**
	 * Returns domain size of a field.
	 * @param field given field.
	 * @return domain size of the specified field.
	 */
	public static int getSizeD(int field) {
		return size_D[field];
	}

	/**
	 * Returns the maximum domain size over all the countable fields.
	 * @return maximum domain size.
	 */
	public static int getMaxSizeD() {
		return maxSize_D;
	}

	/**
	 * Returns number of countable fields.
	 * @return number of countable fields.
	 */
	public static int getNumFields() {
		return numFields;
	}
	
	/**
	 * Returns the total domain size which is the product of domain size of all fields. 
	 * @return total domain size.
	 */
	public static int getTotalSizeD() {
		int size = 1;
		for (int i = 0; i < numFields; i++)
			size *= size_D[i];
		return size;
	}

	static void tic() {
		t = System.currentTimeMillis();
	}

	static void toc() {
		delta_t = System.currentTimeMillis() - t;
	}

	/**
	 * Writes parameters to a configuration file.
	 * @param filename name of the configuration file.
	 */
	public static void writeConfig(String filename) {
		try {
			writeConfig(new FileOutputStream(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Writes parameters to an output stream.
	 * @param os output stream.
	 */
	public static void writeConfig(OutputStream os) {
		PrintWriter pw = new PrintWriter(os);
		pw.println("# Secret");
		pw.println("p=" + p.toString());
		pw.println("b=" + b.toString());
		pw.println();
		pw.println("# Public");
		pw.println("q=" + q.toString());
		pw.println("n=" + n.toString());
		pw.println("s1=" + Integer.toString(s1));
		pw.println("s2=" + Integer.toString(s2));
		pw.println("numFields=" + Integer.toString(numFields));
		for (int i = 0; i < numFields; i++) {
			pw.println("|D" + i + "|=" + Integer.toString(size_D[i]));
		}
		pw.println();
		pw.println("# Statistics");
		pw.println("|p|=" + Integer.toString(p.bitLength()));
		pw.println("|b|=" + Integer.toString(b.bitLength()));
		pw.println("|q|=" + Integer.toString(q.bitLength()));
		pw.println("generate_time=" + Long.toString(delta_t));
		pw.close();
	}

	/**
	 * Parse big integer from a line.
	 * @param line contains "name=value" of a parameter.
	 * @param name name of the parameter.
	 * @param defaultValue default value of the parameter in case parsed value is invalid (or empty). 
	 * @return value of the parameter.
	 */
	private static BigInteger getBigInteger(String line, String name,
			BigInteger defaultValue) {
		BigInteger value = defaultValue;
		line = line.trim();
		if (line.startsWith(name + "=")) {
			value = new BigInteger(line.substring(line.indexOf("=") + 1).trim());
		}

		return value;
	}

	/**
	 * Parse long integer from a line.
	 * @param line
	 * @param name
	 * @param defaultValue
	 * @return
	 */
	private static long getLong(String line, String name, int defaultValue) {
		long value = defaultValue;
		line = line.trim();
		if (line.startsWith(name + "=")) {
			try {
				value = Long.parseLong(line.substring(line.indexOf("=") + 1)
						.trim());
			} catch (NumberFormatException e) {
			}
		}

		return value;
	}

	/**
	 * Reads parameters from a configuration file.
	 * @param filename name of the configuration file.
	 */
	public static void readConfig(String filename) {
		try {
			readConfig(new FileInputStream(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Reads parameters from an input stream.
	 * @param is input stream.
	 */
	@SuppressWarnings("deprecation")
	public static void readConfig(InputStream is) {
		size_D = null;
		maxSize_D = 0;

		try {
			DataInputStream dis = new DataInputStream(is);
			while (dis.available() > 0) {
				String line = dis.readLine().trim();
				p = getBigInteger(line, "p", p);
				b = getBigInteger(line, "b", b);
				q = getBigInteger(line, "q", q);
				n = getBigInteger(line, "n", n);
				s1 = (int) getLong(line, "s1", s1);
				s2 = (int) getLong(line, "s2", s2);
				numFields = (int) getLong(line, "numFields", numFields);

				if (line.startsWith("|D")) {
					if (size_D == null) {
						size_D = new int[numFields];
						for (int i = 0; i < numFields; i++)
							size_D[i] = 0;
					}

					int i = Integer.parseInt(line.substring(2,
							line.indexOf("|=")));
					size_D[i] = Integer.parseInt(line.substring(line
							.indexOf("=") + 1));
					if (size_D[i] > maxSize_D)
						maxSize_D = size_D[i];
				}
			}
			dis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length != 5) {
			System.out.println("Generating parameters and write to a file.");
			System.out
					.println("Usage: java Parameters <n> <size_D> <s1> <s2> <cfgfile>");
			System.out.println("    n         number of records.");
			System.out
					.println("    size_D    array of size of countable fields, separated by commas.");
			System.out.println("    s1        security parameter s1.");
			System.out
					.println("    s2        security parameter s2, number of bits of a random number r.");
			System.out
					.println("    cfgfile   name of output file containing the generated paramters.");
			System.exit(1);
		}

		n = new BigInteger(args[0]);
		String[] str = args[1].split(",");
		numFields = str.length;
		size_D = new int[numFields];
		maxSize_D = 0;
		for (int i = 0; i < numFields; i++) {
			size_D[i] = Integer.parseInt(str[i]);
			if (size_D[i] > maxSize_D)
				maxSize_D = size_D[i];
		}
		s1 = Integer.parseInt(args[2]);
		s2 = Integer.parseInt(args[3]);

		// generate eta, p, b
		GenPrime.generatePrime();

		writeConfig(args[4]);
	}

}
