package common;

import java.math.BigInteger;

/**
 * Provides an implementation for a univariate polynomial <code>P(x)</code>.
 * 
 * @author vohuudtr
 *
 */
public class Polynomial {
	private BigInteger[] coefficients;
	private int deg = 0;

	/**
	 * Constructing the univariate polynomial with a given highest degree that the polynomial can have.
	 * At initialization, all coefficients are set to 0, and the polynomial degree is 0.
	 * 
	 * @param highestDegree the highest degree.
	 */
	public Polynomial(int highestDegree) {
		coefficients = new BigInteger[highestDegree + 1];
		for (int i = 0; i <= highestDegree; i++) {
			coefficients[i] = new BigInteger("0");
		}
		this.deg = 0;
	}

	/**
	 * Constructing the univariate polynomial with given coefficients in Java {@link BigInteger} type.
	 * The polynomial degree is determined based on the values of the given coefficients.
	 * 
	 * @param coefficients given coefficients.
	 */
	public Polynomial(BigInteger[] coefficients) {
		this.coefficients = new BigInteger[coefficients.length];
		for (int i = 0; i < coefficients.length; i++) {
			this.coefficients[i] = new BigInteger(coefficients[i].toByteArray());
			if (this.coefficients[i].signum() != 0)
				this.deg = i;
		}
	}

	/**
	 * Constructing the univariate polynomial with given coefficients in integer type.
	 * The polynomial degree is determined based on the values of the given coefficients.
	 * 
	 * @param coefficients given coefficients.
	 */
	public Polynomial(int[] coefficients) {
		this.coefficients = new BigInteger[coefficients.length];
		for (int i = 0; i < coefficients.length; i++) {
			this.coefficients[i] = new BigInteger(
					Integer.toString(coefficients[i]));
			if (this.coefficients[i].signum() != 0)
				this.deg = i;
		}
	}

	/**
	 * Returns the polynomial degree.
	 * 
	 * @return the polynomial degree.
	 */
	public int degree() {
		return deg;
	}

	/**
	 * Returns coefficient corresponding the given degree.
	 * 
	 * @param degree degree of the queried coefficient.
	 * @return the queried coefficient. If the queried degree is greater than the
	 * polynomial degree, 0 is returned.
	 */
	public BigInteger getCoefficient(int degree) {
		if (degree <= deg)
			return coefficients[degree];
		else
			return new BigInteger("0");
	}

	/**
	 * Returns all coefficients of the polynomial.
	 * @return array of all coefficients.
	 */
	public BigInteger[] getCoefficients() {
		return coefficients;
	}

	/**
	 * Sets coefficient at a specified degree with given value in Java {@link BigInteger} type.
	 * 
	 * @param degree specified degree.
	 * @param coeff new value for the coefficient.
	 */
	public void setCoefficient(int degree, BigInteger coeff) {
		if (degree <= deg)
			coefficients[degree] = coeff;
	}

	/**
	 * Sets coefficient of a specified degree with given value in integer type.
	 * 
	 * @param degree specified degree.
	 * @param coeff new value for the coefficient.
	 */
	public void setCoefficient(int degree, int coeff) {
		if (degree <= deg) {
			coefficients[degree] = new BigInteger(Integer.toString(coeff));
		}
	}

	/**
	 * Evaluates the polynomial at a given point by computing the monomials
	 * with degree greater than or equal to the given <code>startDegree</code>
	 * and add them together.
	 * 
	 * @param x evaluated point.
	 * @param startDegree starting degree that monomials are computed.
	 * @return \f$\sum_{i=0}^{startDegree}a_i\cdot{}x^i\f$
	 */
	private BigInteger value(BigInteger x, int startDegree) {
		if (startDegree <= deg)
			return coefficients[startDegree].add(x.multiply(value(x,
					startDegree + 1)));
		else
			return new BigInteger("0");
	}

	/**
	 * Evaluates the polynomial at a given point.
	 * 
	 * @param x given value <code>x</code>.
	 * @return <code>P(x)</code>.
	 */
	public BigInteger value(BigInteger x) {
		return value(x, 0);
	}

	/**
	 * Returns a new univariate polynomial as the sum
	 * of this polynomial and another polynomial.
	 * 
	 * @param p another polynomial.
	 * @return sum of the polynomials.
	 */
	public Polynomial add(Polynomial p) {
		int highest = p.deg < this.deg ? this.deg
				: p.deg;
		Polynomial q = new Polynomial(highest);
		for (int i = 0; i <= highest; i++) {
			q.coefficients[i] = this.getCoefficient(i).add(p.getCoefficient(i));
			if (q.coefficients[i].signum() != 0)
				q.deg = i;
		}

		return q;
	}

	/**
	 * Returns a new univariate polynomial as the result of
	 * subtracting this polynomial by another polynomial.
	 * 
	 * @param p another polynomial.
	 * @return subtraction of this polynomial by another polynomial.
	 */
	public Polynomial subtract(Polynomial p) {
		int highest = p.deg < this.deg ? this.deg
				: p.deg;
		Polynomial q = new Polynomial(highest);
		for (int i = 0; i <= highest; i++) {
			q.coefficients[i] = this.getCoefficient(i).subtract(
					p.getCoefficient(i));
			if (q.coefficients[i].signum() != 0)
				q.deg = i;
		}

		return q;
	}

	/**
	 * Returns a new univariate polynomial by negating this polynomial.
	 * 
	 * @return the negation of this polynomial.
	 */
	public Polynomial negate() {
		Polynomial q = new Polynomial(this.deg);
		for (int i = 0; i <= this.deg; i++) {
			q.coefficients[i] = this.coefficients[i].negate();
		}
		q.deg = this.deg;

		return q;
	}

	/**
	 * Returns a new univariate polynomial by multiplying this polynomial with a scalar value in Java BigInteger type.
	 * 
	 * @param k scalar value.
	 * @return result of the multiplication.
	 */
	public Polynomial multiply(BigInteger k) {
		Polynomial q = new Polynomial(this.deg);
		for (int i = 0; i <= this.deg; i++) {
			q.coefficients[i] = this.coefficients[i].multiply(k);
		}
		if (k.signum() != 0)
			q.deg = this.deg;
		else
			q.deg = 0;

		return q;
	}

	/**
	 * Returns a new univariate polynomial by multiplying this polynomial with a scalar value in integer type.
	 * 
	 * @param k scalar value.
	 * @return result of the multiplication.
	 */
	public Polynomial multiply(int k) {
		Polynomial q = new Polynomial(this.deg);
		for (int i = 0; i <= this.deg; i++) {
			q.coefficients[i] = this.coefficients[i].multiply(new BigInteger(
					Integer.toString(k)));
		}
		if (k != 0)
			q.deg = this.deg;
		else
			q.deg = 0;

		return q;
	}

	/**
	 * Returns a new univariate polynomial by multiplying this polynomial with another polynomial.
	 * The multiplication is done among coefficients of two polynomials.
	 * 
	 * 
	 * @param p another polynomial.
	 * @return result of the multiplication.
	 */
	public Polynomial multiply(Polynomial p) {
		Polynomial q = new Polynomial(this.deg + p.deg);
		for (int i = 0; i < q.coefficients.length; i++) {
			for (int k = 0; k <= i; k++) {
				if (k < this.coefficients.length
						&& i - k < p.coefficients.length) {
					q.coefficients[i] = q.coefficients[i]
							.add(this.coefficients[k].multiply(p.coefficients[i
									- k]));
				}
			}
			if (q.coefficients[i].signum() != 0)
				q.deg = i;
		}

		return q;
	}

	/**
	 * Returns a new univariate polynomial as an exponentiation of this polynomial.
	 * 
	 * @param exponent desired exponent.
	 * @return result of the exponentiation.
	 */
	public Polynomial pow(int exponent) {
		if (exponent == 0)
			return new Polynomial(new int[] { 1 });
		else if (exponent == 1)
			return new Polynomial(this.coefficients);
		else {
			// fast power by splitting into 2 parts and multiply them together
			if (exponent % 2 == 0) {
				Polynomial p = this.pow(exponent / 2);
				return p.multiply(p);
			} else {
				Polynomial p = this.pow(exponent / 2);
				return this.multiply(p.multiply(p));
			}
		}
	}

	/**
	 * Returns a new univariate polynomial by mod-ing all coefficients of this polynomial by a given modulus.
	 * @param m the modulus.
	 * @return result of the mod operation.
	 */
	public Polynomial mod(BigInteger m) {
		Polynomial q = new Polynomial(this.deg);
		for (int i = 0; i <= this.deg; i++) {
			q.coefficients[i] = this.coefficients[i].mod(m);
			if (q.coefficients[i].signum() != 0)
				q.deg = i;
		}

		return q;
	}

	/**
	 * Returns a string representing this polynomial.
	 */
	@Override
	public String toString() {
		String s = new String("[");

		for (int i = 0; i <= deg; i++)
			if (i == 0)
				s = s + coefficients[i].toString();
			else
				s = s + " " + coefficients[i].toString();
		s = s + "]";

		return s;
	}

	/**
	 * Creates a univariate polynomial in Lagrange form.
	 * 
	 * \f$l_i(x) = \prod_{m \ne i} \frac{x-x_m}{x_i-x_m}\f$
	 * 
	 * @param x array of all points \f$x_m\f$.
	 * @param i index \f$i\f$ of \f$x_i\f$.
	 * 
	 */
	public static Polynomial createLagrange(int[] x, int i) {
		Polynomial p = new Polynomial(x.length - 1);

		p.setCoefficient(0, 1);
		for (int m = 0; m < x.length; m++) {
			if (m != i) {
				p = p.multiply(new Polynomial(new int[] { -x[m], 1 }));
			}
		}

		p = p.multiply(
				p.value(new BigInteger(Integer.toString(x[i]))).modInverse(
						Parameters.getQ())).mod(Parameters.getQ());
		
		return p;
	}

	/**
	 * For testing purpose.
	 * @param args
	 */
	public static void main(String args[]) {
		Polynomial p = new Polynomial(new int[] { 1, 1 });
		Polynomial q = new Polynomial(new int[] { -2, 1 });
		Polynomial r = p.subtract(q);
		System.out.println(p.value(new BigInteger("10000")));
		System.out.println(r.toString());
	}
}
