package common;

import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.math.BigInteger;
import java.util.Random;

import mapred.BigIntegerWritable;

/**
 * Implementation of a local data generator. This class is used for experiments.
 * <p>
 * A record is formatted in order of columns as follows:
 * <ul>
 * <li>Encrypted countable field 1.</li>
 * <li>Plain-text countable field 1.</li>
 * <li>Encrypted countable field 2.</li>
 * <li>Plain-text countable field 2.</li>
 * ...
 * <li>Encrypted countable field m.</li>
 * <li>Plaini-text countable field m.</li>
 * <li>Remaining random data</li>
 * </ul>
 * Notes:
 * <ul>
 * <li>Size of each record is specified in {@link Record}.</li>
 * <li>Number of countable fields, m, is specified in the parameters file (see {@link Parameters}).</li>
 * <li>Size of each encrypted countable field depends on size of prime p (see {@link Parameters}).
 * Encrypted countable field's format is defined in {@link BigIntegerWritable} class.</li>
 * </ul>
 * 
 * @see #main(String[]) for detailed usage.
 * 
 * @author vohuudtr
 *
 */
public class Producer {
	public static OutputStream paramos = null;
	public static OutputStream statos = null;
	private static Statistics stat = new Statistics();
	private static String datafile = "";
	private static long sizePerFile = 1000000000;
	private static Random random = new Random(System.currentTimeMillis());

	/**
	 * Generate a random record. While generating records, statistics about the values
	 * are also recorded for evaluation.
	 * 
	 * @return the generated record.
	 */
	public static Record generateRecord() {
		Record record = new Record(Parameters.getNumFields() * 2);
		for (int j = 0; j < Parameters.getNumFields(); j++) {
			// generate a random number
			int random_x = random.nextInt(Parameters.getSizeD(j));
			BigIntegerWritable x = new BigIntegerWritable(
					Cipher.encrypt(new BigInteger(Integer.toString(random_x))));
			record.setField(j * 2, x);
			record.setField(j * 2 + 1, new BigIntegerWritable(random_x));
		}

		// record statistics
		stat.add(record);

		return record;
	}

	/**
	 * Returns the statistics of generated records.
	 * 
	 * @return statistics of generated records.
	 */
	public static Statistics getStats() {
		return stat;
	}

	/**
	 * Generate "quantity" random numbers with values between 0 (inclusive) and maxValues (inclusive).
	 * The "maxValues" is specified in the parameter file.
	 * @param quantity number of generated random numbers.
	 * @throws IOException exception.
	 */
	private static void generate(long quantity, OutputStream os, boolean split)
			throws IOException {
		DataOutputStream dos = null;
		long i;
		long curSize = 0;
		long totalSize = 0;
		int fileNum = 0;
		int percent = 0;
		long time = System.currentTimeMillis();

		totalSize = quantity * Record.blocksize;

		if (os == null && split == false) {
			os = System.out;
		}

		if (os != null) {
			dos = new DataOutputStream(os);
		}

		// then generate and write generated E(x) to data file
		for (i = 0; i < quantity; i++) {
			if (split) {
				if (dos == null || curSize > sizePerFile) {
					if (dos != null) {
						System.err.println("Written to " + datafile
								+ Integer.toString(fileNum));
						dos.close();
						os.close();
					}
					fileNum++;

					os = new FileOutputStream(datafile
							+ Integer.toString(fileNum));
					dos = new DataOutputStream(os);
				}
			}

			generateRecord().write(dos);

			curSize += Record.blocksize;

			if (curSize * 100 / totalSize > percent) {
				percent = (int) (curSize * 100 / totalSize);
				System.err.println("Written " + curSize + " out of "
						+ totalSize + " bytes..." + percent + "%");
			}
		}

		if (dos != null) {
			dos.close();
			os.close();
		}

		System.err.println((System.currentTimeMillis() - time) + "ms");

		// finally, put statistics into stat file
		if (statos != null)
			stat.print(statos);
	}

	private static void usage() {
		System.err
				.println("Generating random numbers between 0 (inclusive) and max (inclusive)");
		System.err.println("Usage: java " + Producer.class.getName()
				+ " <quantity> <paramfile> <statfile> [datafile]");
		System.err
				.println("   quantity    Specifies number of generated ciphertext.");
		System.err
				.println("   paramfile   Specifies location of the parameter file.");
		System.err
				.println("   statfile    Specifies location of the file for statistics.");
		System.err
				.println("   datafile    Specifies location of the file storing the generated numbers.");
		System.err
				.println("               If not specified, standard output is used.");
		System.err
				.println("Example: The following command produces 1000 numbers using params.txt parameter file:");
		System.err
				.println("   java Producer 1000 params.txt stat-orig-data.txt data.txt");
	}

	static void generate(long quantity) throws IOException {
		generate(quantity, null, false);
	}

	static void generate(long quantity, boolean split)
			throws IOException {
		generate(quantity, null, split);
	}

	static void generate(long quantity, OutputStream os)
			throws IOException {
		generate(quantity, os, false);
	}

	/**
	 * Usage:
	 * <p>
	 * <code>java common.Producer &lt;quantity&gt; &lt;paramfile&gt; &lt;statfile&gt; [datafile]</code>
	 * <dl>
	 * <dt>quantity</dt><dd>Number of records to generate.</dd>
	 * <dt>paramfile</dt><dd>Path to the parameters file (see {@link Parameters}).</dd>
	 * <dt>statfile</dt><dd>Path to the output statistics file.</dd>
	 * <dt>datafile</dt><dd>Path to the output data file. The output is automatically split into multiple
	 * files (suffixed with numbers) of size not greater than 1GB.
	 * If <code>datafile</code> is not specified, the standard output is used.</dd>
	 * </dl>
	 * All the paths can be either absolute or relative paths.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length >= 3) {
			try {
				long quantity = Long.parseLong(args[0]);
				boolean split = false;

				statos = new FileOutputStream(args[2]);
				if (args.length >= 4) {
					datafile = args[3];
					split = true;
				}

				Cipher.initialize(args[1]);
				stat.initialize(Parameters.getTotalSizeD());
				generate(quantity, null, split);

				System.exit(0);
			} catch (NumberFormatException e) {
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		usage();
		System.exit(1);
	}

}
