package common;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

import mapred.BigIntegerWritable;

import org.apache.hadoop.io.Writable;

/**
 * Definition of a record used by the application.
 * <p>
 * Each record has a fixed size that can be changed via {@link #blocksize}.
 * <p>
 * Each record also has a fixed number of fields.
 * <p>
 * <em>Note:</em> for convenient experiments, a generated record contains
 * twice the number of countable fields defined in the parameter file,
 * i.e. if the number of countable fields is m, the actual number of fields
 * defined in each record is 2m, where an even-indexed 2k-th field contains the
 * encrypted value of the k-th field, while an odd-indexed (2k+1)-th field contains the
 * plain-text value of the k-th field.
 * 
 * @author vohuudtr
 *
 */
public class Record implements Writable {
	/**
	 * Specifies a fixed size in bytes for each record.
	 */
	public static int blocksize = 1000000;

	private BigIntegerWritable[] fields;
	private int numFields = 0;

	/**
	 * Initializes the record with zero fields.
	 */
	public Record() {
		fields = new BigIntegerWritable[numFields];
	}

	/**
	 * Initializes the record with a given number of fields.
	 * @param numFields number of fields.
	 */
	public Record(int numFields) {
		this.numFields = numFields;
		fields = new BigIntegerWritable[numFields];
	}

	/**
	 * Re-initializes the record by reading from the input stream.
	 * <p>
	 * Format of record in the input is as follows:
	 * <ul>
	 * <li>First 4 bytes represent an integer, which specifies the number of fields (2m).</li>
	 * <li>Field 1 (i.e. encrypted field 1).</li>
	 * <li>Field 2 (i.e. plain-text field 1).</li>
	 * <li>Field 3 (i.e. encrypted field 2).</li>
	 * <li>Field 4 (i.e. plain-text field 2).<li>
	 * ...
	 * <li>Field 2m (i.e. encrypted field m).</li>
	 * <li>Field 2m+1 (i.e. plain-text field m).</li>
	 * <li>Remaining random bytes for other fields that are not considered.</li>
	 * </ul>
	 * All fields are in {@link BigIntegerWritable} format. 
	 */
	@Override
	public void readFields(DataInput in) throws IOException {
		numFields = in.readInt();
		fields = new BigIntegerWritable[numFields];
		for (int i = 0; i < numFields; i++)
			fields[i] = new BigIntegerWritable(in);
		in.skipBytes(blocksize - getSize());
	}

	/**
	 * Writes the record to the given output stream using the format specified in {@link #readFields(DataInput)}.
	 */
	@Override
	public void write(DataOutput out) throws IOException {
		out.writeInt(numFields);
		for (int i = 0; i < numFields; i++)
			fields[i].write(out);
		out.write(new byte[blocksize - getSize()]);
	}

	/**
	 * Returns total size in bytes of all fields in the record, not including the rest of random bytes.
	 * @return total size.
	 */
	public int getSize() {
		int size = 4;

		for (int i = 0; i < numFields; i++)
			size += fields[i].getSize();

		return size;
	}

	/**
	 * Return values of the specified field.
	 * @param i index of the field.
	 * @return value of the field.
	 */
	public BigIntegerWritable getField(int i) {
		return fields[i];
	}

	/**
	 * Set value of the specified field.
	 * @param i index of the field.
	 * @param value value of the field.
	 */
	public void setField(int i, BigIntegerWritable value) {
		fields[i] = value;
	}

	/**
	 * Returns the number of fields in the record (not considering random bytes of the rest).
	 * @return number of fields.
	 */
	public int getNumFields() {
		return numFields;
	}
}
