package mapred;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.math.BigInteger;

import org.apache.hadoop.io.Writable;

/**
 * Provides an <em>immutable</em> implementation of a big integer for use in Hadoop framework.
 * Similarly to the Java {@link BigInteger} class, this class supports two basic mathematical operations: <b>addition</b> and <b>multiplication</b>.
 * Besides, it also implements the {@link Writable} interface to support <b>reading</b> and <b>writing</b> operations in Hadoop framework.
 * 
 * 
 * @author vohuudtr
 *
 */
public class BigIntegerWritable implements Writable {
	private byte[] bytes;
	private BigInteger value;
	private int size = 0;

	/**
	 * Default constructor without initialization. This constructor is needed for the iterator used in the reduce method of <code>MapRed</code>-classes.
	 * Therefore, one should <em>manually</em> initialize the value of this object after using this constructor.
	 */
	public BigIntegerWritable() {
	}

	/**
	 * Constructs a BigIntegerWritable object from an integer.
	 * @param n initialized value for the constructed object.
	 */
	public BigIntegerWritable(int n) {
		value = new BigInteger(Integer.toString(n));
	}

	/**
	 * Constructs a BigIntegerWritable object from a Java BigInteger object.
	 * @param n initialized value for the constructed object.
	 */
	public BigIntegerWritable(BigInteger n) {
		value = n;
	}

	/**
	 * Constructs a BigIntegerWritable object from the input stream. This constructor
	 * is a short-hand for initializing the object with the default constructor
	 * and calling {@link #readFields(DataInput)} to read the value from the input stream.
	 * @param in input stream
	 * @throws IOException if IO errors occur.
	 * @see #readFields(DataInput) for storing format of the object in the stream.
	 */
	public BigIntegerWritable(DataInput in) throws IOException {
		readFields(in);
	}

	/**
	 * Initializes the object by reading data from the input stream.
	 * <p>
	 * The object is stored in a stream as an <em>ordered sequence</em> of bytes:
	 * <ul>
	 * <li>First 4 bytes represents an integer which specifies the number of bytes
	 * used for storing the "big-integer" value of this object.</li>
	 * <li>The value of the object as an array of bytes with the length specified
	 * in the above 4 bytes. The bytes storing order depends on the Java
	 * BigInteger implementation.</li> 
	 * </ul>
	 * 
	 * @param in input stream.
	 * @throws IOException if IO errors occur.
	 */
	@Override
	public void readFields(DataInput in) throws IOException {
		bytes = new byte[in.readInt()];
		in.readFully(bytes);
		value = new BigInteger(bytes);

		size = 4 + bytes.length;
	}

	/**
	 * Writes the object into the specified output stream.
	 * The sequence of bytes representing the object is in the format described in {@link #readFields(DataInput)}.
	 * 
	 * @param out output stream.
	 * @throws IOException if IO errors occur.
	 * @see #readFields(DataInput) for storing format of the object in the stream.
	 */
	@Override
	public void write(DataOutput out) throws IOException {
		if (size == 0) {
			serialize();
		}

		out.writeInt(bytes.length);
		out.write(bytes);
	}

	/**
	 * Returns the total size in bytes representing the value of the object.
	 * @return total size in bytes.
	 */
	public int getSize() {
		if (size == 0)
			serialize();

		return bytes.length + 4;
	}

	private void serialize() {
		if (size == 0) {
			bytes = value.toByteArray();
			size = 4 + bytes.length;
		}
	}

	/**
	 * Returns a new BigIntegerWritable object with value equal to the sum
	 * of this object's value and another object's value.
	 * @param a another object.
	 * @return result of the addition.
	 */
	public BigIntegerWritable add(BigIntegerWritable a) {
		return new BigIntegerWritable(value.add(a.value));
	}

	/**
	 * Returns a new BigIntegerWritable object with value equal to the product
	 * of this object's value and another object's value.
	 * @param a another object.
	 * @return result of the multiplication.
	 */
	public BigIntegerWritable multiply(BigIntegerWritable a) {
		return new BigIntegerWritable(value.multiply(a.value));
	}

	/**
	 * Returns a string representing this object.
	 */
	@Override
	public String toString() {
		return value.toString();
	}

}
