package mapred;

import java.io.DataInput;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.net.URI;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.hdfs.DistributedFileSystem;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import common.Cipher;
import common.Parameters;

/**
 * EPiC MapReduce main class. Based on provided parameters via command line,
 * an appropriate MapReduce class is executed. Currently, there are two
 * implementation of counting job that can be called via Count:
 * <dl>
 * <dt>{@link MapRedEpic}</dt><dd>EPiC approach - counting on encrypted fields.</dd>
 * <dt>{@link MapRedPlainCountOne}</dt><dd>Plain-text counting.</dd>
 * </dl>
 * <p>
 * Usage of Count via command-line:
 * <p>
 * <code>hadoop jar &lt;JARFILE&gt; %mapred.Count [options] &lt;input&gt; &lt;output&gt;</code>
 * <p>
 * <dl>
 * <dt>JARFILE</dt>
 * <dd>The JAR file containing EPiC.</dd>
 * <dt>input</dt>
 * <dd>A HDFS path to the directory containing the input data.<br>
 * <em>Note:</em> Only input files in the specified directory which start with "data" are read. 
 * </dd>
 * <dt>output</dt>
 * <dd>A HDFS path to the directory containing the results.<br>
 * <em>Note:</em> Existing output will be <em>automatically</em> removed when Count is started.
 * </dd>
 * </dl>
 * 
 * The options given to Hadoop must be prefixed by "-D". The following options are supported:
 * <dl>
 * <dt>paramfile</dt>
 * <dd>Name of the encryption parameters file. Path to the parameters file must be relative to
 * this class inside the the JAR file.<br>
 *               Example: <code>-Dparamfile=params.txt</code>
 * </dd>
 * <dt>request</dt>
 * <dd>A HDFS path to the user's request file containing the counting query. The specified
 *               path is relative to the <b>input</b> directory.<br>
 *               Example: <code>-Drequest=request</code>
 * </dd>
 * <dt>mapred</dt>
 * <dd>Specifies which MapReduce approach to be executed. Currently the following values are supported:
 * <dl>
 * <dt>epic</dt><dd>Calling {@link MapRedEpic} to execute the query, which applies a variant version of the approach
 * presented in the EPiC paper. Precisely, based on the provided <code>request</code>,
 * which contains the encrypted coefficients of the queried indicator polynomial, the Mappers evaluate the indicator polynomial
 * for each record by multiplying the monomials with the coefficients before adding them together. In the last step at the Reducer,
 * those results from Mappers (now considered as the value of the indicator polynomial evaluated for the corresponding subsets)
 * are added together to obtain the final results and return to the user.
 * <p>The approach presented in the EPiC paper is implemented in {@link MapRedEpicReducerEvaluate}, in which the Mappers compute
 * the monomials without multiplying with the coefficients. At the final step, the Reducer adds those results from the Mappers
 * together and then multiplies with the given coefficients to yield the final results.
 * <p>An older approach of EPiC (see {@link MapRedNotSendCoeff}) is to keep the coefficients at the user side. The Mappers and Reducers only need to compute
 * the monomials and add them together, then return to the user, who will be responsible to multiply the results with
 * the precomputed coefficients to obtain the counting result. This, however, requires much more communication for downloading
 * the results, therefore, is impractical. 
 * </dd>
 * <dt>plain</dt><dd>Calling {@link MapRedPlainCountOne} to count on plain-text values of the multiple countable fields.
 * This implementation does not support range counting, boolean expressions, etc. Another illustration of plain-text counting
 * is implemented in {@link MapRedPlainCountAll} which counts all possible values in one MapReduce job.</dd>
 * </dl>
 * Example: <code>-Dmapred=epic</code>
 * </dd>
 * </dl> 
 * @author vohuudtr
 *
 */
public class Count extends Configured implements Tool {

	/**
	 * Returns the data input interface for a file with specified patterns among given file paths.
	 * This method should be used inside the package only.
	 * 
	 * @param files set of files.
	 * @param name pattern that ends the files.
	 * @return the data input interface corresponding to the matched file.
	 */
	static DataInput getInput(Path[] files, String name) {
		for (int i = 0; i < files.length; i++) {
			if (files[i].toString().endsWith(name)) {
				FileInputStream fis;
				try {
					fis = new FileInputStream(files[i].toString());
					DataInputStream dis = new DataInputStream(fis);
					return dis;
				} catch (FileNotFoundException e) {
					e.printStackTrace();
				}
				return null;
			}
		}

		return null;
	}

	/**
	 * Run the job.
	 * 
	 * @param args argument list for the running job.
	 * @throws Exception if errors occur.
	 */
	public int run(String[] args) throws Exception {
		JobConf conf = new JobConf(getConf(), Count.class);

		DistributedFileSystem fs = (DistributedFileSystem) FileSystem.get(conf);
		// if paramfile is defined, read file from JAR, and set parameters for mappers/reducers
		if (conf.get("paramfile") != null) {
			Cipher.initialize(Count.class.getResourceAsStream(conf
					.get("paramfile")));

			// set conf parameters for mappers and reducers
			for (int i = 0; i < Parameters.getNumFields(); i++) {
				conf.setInt("|D" + i + "|", Parameters.getSizeD(i));
			}
			conf.setInt("numFields", Parameters.getNumFields());
		}

		conf.set("request", fs.getHomeDirectory() + "/" + args[0] + "/request");
		DistributedCache.addCacheFile(new URI(conf.get("request")), conf);
		String mapred = conf.get("mapred");
		if (mapred.equals("epic"))
			MapRedEpic.setup(conf);
		else if (mapred.equals("plain"))
			MapRedPlainCountOne.setup(conf);

		// delete old output before running job
		fs.delete(new Path(args[1]), true);

		FileInputFormat.setInputPaths(conf, new Path(args[0] + "/data*"));
		FileOutputFormat.setOutputPath(conf, new Path(args[1]));

		long t = System.currentTimeMillis();
		JobClient.runJob(conf);

		PrintWriter pw = new PrintWriter(new FileWriter("count.txt", true));
		pw.print((System.currentTimeMillis() - t));
		pw.close();

		return 0;
	}

	/**
	 * Entry point of the class.
	 * @param args command-line arguments provided to the class.
	 * @throws Exception if errors occur.
	 */
	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new Count(), args);
		System.exit(res);
	}
}
