package mapred;
import java.io.IOException;

import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapred.FileSplit;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.RecordReader;

import common.Record;

/**
 * Provides an implementation of a common record reader for all MapReduce jobs in the distributed application.
 * Each record has a fixed size, defined in {@link Record}.
 * 
 * @author vohuudtr
 *
 */
public class CustomRecordReader implements RecordReader<LongWritable, Record> {
	private FSDataInputStream fileIn;
	private long start;
	private long end;
	private long pos;
	private int blocksize;

	public CustomRecordReader(JobConf job, FileSplit split)  throws IOException {
		blocksize = Record.blocksize;

		FileSystem fs = split.getPath().getFileSystem(job);
	    fileIn = fs.open(split.getPath());
	    end = split.getStart() + split.getLength();
	    start = split.getStart();
	    // if we are not at the first byte of the block,
	    // move to next block
	    if (start % blocksize != 0)
	    	start += blocksize - start % blocksize;
        pos = start;
        fileIn.seek(start);
	}
	
	@Override
	public void close() throws IOException {
		if (fileIn != null)
			fileIn.close();
	}

	@Override
	public LongWritable createKey() {
		return new LongWritable();
	}

	@Override
	public Record createValue() {
		return new Record();
	}

	@Override
	public long getPos() throws IOException {
		return pos;
	}

	@Override
	public float getProgress() throws IOException {
		if (start == end)
			return 0.0f;
		else
			return Math.min(1.0f, (pos - start) / (float)(end - start));
	}

	@Override
	public boolean next(LongWritable key, Record record) throws IOException {
		if (pos < end) {
			// set current position as key
			key.set(pos);
			
			// read record
			record.readFields(fileIn);
			
			pos += blocksize;
			
			return true;
		}
		
		return false;
	}
	
}

