package mapred;

import java.io.DataInput;
import java.io.IOException;
import java.util.Iterator;

import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextOutputFormat;

import common.Cipher;
import common.MultiArray;
import common.MultiFieldKey;
import common.Record;

/**
 * Implementation of EPiC's approach. The Mappers evaluate the indicator polynomial
 * for the Mappers' input split by computing the monomials for each record and multiplying them
 * with the encrypted coefficients of the given polynomial in the user's request. Results for each
 * record in the subset are added together and sent to the Reducer. The Reducer simply adds the results
 * from Mappers together and obtain the final results for the whole data set.
 * <p>
 * This implementation supports all kinds of counting (conjunctive, disjunctive, range, GF(2) arithmetized).
 * 
 * @author vohuudtr
 * 
 */
public class MapRedEpic {

	private static class CountMapper extends MapReduceBase implements
			Mapper<LongWritable, Record, NullWritable, BigIntegerWritable> {
		private MultiArray<BigIntegerWritable> sum;
		private MultiArray<BigIntegerWritable> coeffs;
		private OutputCollector<NullWritable, BigIntegerWritable> output;
		private int numFields;
		private int[] size_D;
		private Path[] sharedFiles;

		public void configure(JobConf conf) {
			Cipher.initialize(Count.class.getResourceAsStream(conf
					.get("paramfile")));
			try {
				sharedFiles = DistributedCache.getLocalCacheFiles(conf);
				DataInput request = Count.getInput(sharedFiles, "request");
				if (request == null)
					throw new RuntimeException("No coefficients found.");
				else {
					coeffs = MultiArray
							.readMultiArrayBigIntegerWritable(request);
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
			numFields = conf.getInt("numFields", 0);
			size_D = new int[numFields];
			for (int i = 0; i < numFields; i++) {
				size_D[i] = conf.getInt("|D" + i + "|", 0);
			}
			sum = MultiArray.createBigIntegerWritableArray(coeffs.getSize());
		}

		public void map(LongWritable key, Record record,
				OutputCollector<NullWritable, BigIntegerWritable> output,
				Reporter reporter) throws IOException {
			this.output = output;
			computeSum(0, new MultiFieldKey(numFields), new BigIntegerWritable(
					0), record, new BigIntegerWritable(1));
		}

		/**
		 * Recursively compute x_0^{i_0} * x_1^{i_1} * .... for every
		 * i_0=0,...,size_D[0]-1 ; i_1=0,...,size_D[1]-1 ; ...
		 * 
		 * @param field
		 * @param key
		 * @param value
		 * @param record
		 */
		public void computeSum(int field, MultiFieldKey key,
				BigIntegerWritable value, Record record,
				BigIntegerWritable runningProd) {
			BigIntegerWritable c;
			c = record.getField(field * 2);

			for (int v = 0; v < size_D[field]; v++) {
				key = key.clone();
				key.element[field] = v;
				if (field == numFields - 1) {
					BigIntegerWritable lastsum = sum.get(key);
					if (lastsum == null)
						lastsum = new BigIntegerWritable(0);
					sum.put(key, lastsum.add(runningProd));
				} else {
					computeSum(field + 1, key, value, record, runningProd);
				}
				runningProd = runningProd.multiply(c);
			}
		}

		public void writeSum() {
			BigIntegerWritable count = new BigIntegerWritable(0);
			for (int i = 0; i < coeffs.getSize(); i++) {
				count = count.add(coeffs.get(i).multiply(sum.get(i)));
			}

			try {
				output.collect(NullWritable.get(), count);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		public void close() {
			writeSum();
		}
	}

	private static class CountReducer extends MapReduceBase
			implements
			Reducer<NullWritable, BigIntegerWritable, NullWritable, BigIntegerWritable> {
		private OutputCollector<NullWritable, BigIntegerWritable> output;
		private BigIntegerWritable sum;

		public void configure(JobConf conf) {
			sum = new BigIntegerWritable(0);
		}

		public void reduce(NullWritable key,
				Iterator<BigIntegerWritable> values,
				OutputCollector<NullWritable, BigIntegerWritable> output,
				Reporter reporter) throws IOException {
			this.output = output;

			while (values.hasNext()) {
				BigIntegerWritable value = values.next();
				sum = sum.add(value);
			}
		}

		public void close() {
			try {
				output.collect(NullWritable.get(), sum);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Initializes the class based on provided configuration parameters.
	 * The initialization comprises setting job's name, assigning Mapper
	 * and Reducer class as well as their input and output key-value classes.
	 * @param conf configuration parameters.
	 */
	public static void setup(JobConf conf) {
		conf.setJobName("epic");
		conf.setNumReduceTasks(1);
		conf.setMapperClass(CountMapper.class);
		conf.setReducerClass(CountReducer.class);
		conf.setMapOutputKeyClass(NullWritable.class);
		conf.setMapOutputValueClass(BigIntegerWritable.class);
		conf.setOutputKeyClass(NullWritable.class);
		conf.setOutputValueClass(BigIntegerWritable.class);

		conf.setInputFormat(RecordInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);

	}

}
