package mapred;

import java.io.IOException;
import java.util.Iterator;

import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextOutputFormat;

import common.Cipher;
import common.MultiArray;
import common.MultiFieldKey;
import common.Record;

/**
 * This is an older implementation of EPiC. The Mappers and Reducer only compute the
 * monomials and return back to the user. The user is responsible for reconstructing
 * the count value by multiplying the obtained results with the precomputed coefficients.
 * 
 * @author vohuudtr
 * 
 */
public class MapRedNotSendCoeff {

	private static class CountMapper extends MapReduceBase implements
			Mapper<LongWritable, Record, MultiFieldKey, BigIntegerWritable> {
		private MultiArray<BigIntegerWritable> sum;
		private OutputCollector<MultiFieldKey, BigIntegerWritable> output;
		private int numFields;
		private int[] size_D;
		private boolean encrypted;

		public void configure(JobConf conf) {
			Cipher.initialize(Count.class.getResourceAsStream(conf
					.get("paramfile")));

			numFields = conf.getInt("numFields", 0);
			size_D = new int[numFields];
			int capacity = 1;
			for (int i = 0; i < numFields; i++) {
				size_D[i] = conf.getInt("|D" + i + "|", 0);
				capacity *= size_D[i];
			}
			encrypted = conf.getBoolean("encrypted", true);
			sum = MultiArray.createBigIntegerWritableArray(capacity);
		}

		public void map(LongWritable key, Record record,
				OutputCollector<MultiFieldKey, BigIntegerWritable> output,
				Reporter reporter) throws IOException {
			this.output = output;
			if (!encrypted)
				computeSumUnencrypted(record);
			else
				computeSum(0, new MultiFieldKey(numFields),
						new BigIntegerWritable(0), record,
						new BigIntegerWritable(1));
		}

		public void computeSumUnencrypted(Record record) {
			MultiFieldKey key = new MultiFieldKey(numFields);
			for (int j = 0; j < numFields; j++) {
				int v = Integer.parseInt(record.getField(j * 2 + 1).toString());
				key.element[j] = v;
			}
			BigIntegerWritable lastsum = sum.get(key);
			if (lastsum == null)
				sum.put(key, new BigIntegerWritable(1));
			else
				sum.put(key, lastsum.add(new BigIntegerWritable(1)));
		}

		@SuppressWarnings("unused")
		public int calculateDegree(MultiFieldKey key) {
			int degree = 0;
			for (int i = 0; i < key.element.length; i++)
				degree += key.element[i];
			return degree;
		}

		/**
		 * Recursively compute x_0^{i_0} * x_1^{i_1} * .... for every
		 * i_0=0,...,size_D[0]-1 ; i_1=0,...,size_D[1]-1 ; ...
		 * 
		 * @param field
		 * @param key
		 * @param value
		 * @param record
		 */
		public void computeSum(int field, MultiFieldKey key,
				BigIntegerWritable value, Record record,
				BigIntegerWritable runningProd) {
			BigIntegerWritable c;
			c = record.getField(field * 2);
			// System.out.println("field " + field + " c "
			// + Cipher.decrypt(new BigInteger(c.toString())));

			for (int v = 0; v < size_D[field]; v++) {
				key = key.clone();
				key.element[field] = v;
				// System.out.println("key " + key + " prod " +
				// Cipher.decrypt(new BigInteger(runningProd.toString()),
				// calculateDegree(key)));
				if (field == numFields - 1) {
					BigIntegerWritable lastsum = sum.get(key);
					if (lastsum == null)
						lastsum = new BigIntegerWritable(0);
					sum.put(key, lastsum.add(runningProd));
					// System.out.println(Runtime.getRuntime().freeMemory() +
					// " " + Runtime.getRuntime().totalMemory());
					// try {
					// output.collect(key, runningProd);
					// } catch (IOException e) {
					// e.printStackTrace();
					// }
					// MultiFieldKey k = new MultiFieldKey(4);
					// k.element[0] = 3;
					// k.element[1] = 1;
					// k.element[2] = 1;
					// k.element[3] = 1;
					// if (key.equals(k)) {
					// System.out.print("key "
					// + key + " | ");
					// for (int j = 0; j < 4; j++) {
					// System.out.print(Cipher.decrypt(new
					// BigInteger(record.getField(j * 2).toString())) + " ");
					// }
					// System.out.print(" degree " + calculateDegree(key));
					// System.out.print(" runningProd " + Cipher.decrypt(new
					// BigInteger(runningProd.toString()),
					// calculateDegree(key)));
					// System.out.println(" sum "
					// + Cipher.decrypt(new BigInteger(sum.get(key)
					// .toString()), calculateDegree(key)));
					// }
				} else {
					computeSum(field + 1, key, value, record, runningProd);
				}
				runningProd = runningProd.multiply(c);
			}
		}

		public void writeSum() {
			BigIntegerWritable[] items = sum.getAll();
			for (int i = 0; i < items.length; i++) {
				try {
					if (items[i] != null)
						output.collect(MultiArray.getMultiFieldKey(i), items[i]);
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

		public void close() {
			writeSum();
		}
	}

	private static class CountReducer extends MapReduceBase
			implements
			Reducer<MultiFieldKey, BigIntegerWritable, MultiFieldKey, BigIntegerWritable> {
		public void reduce(MultiFieldKey key,
				Iterator<BigIntegerWritable> values,
				OutputCollector<MultiFieldKey, BigIntegerWritable> output,
				Reporter reporter) throws IOException {
			BigIntegerWritable sum = new BigIntegerWritable(0);
			while (values.hasNext()) {
				BigIntegerWritable value = values.next();
				sum = sum.add(value);
			}
			output.collect(key, sum);
		}
	}

	/**
	 * Initializes the class based on provided configuration parameters.
	 * The initialization comprises setting job's name, assigning Mapper
	 * and Reducer class as well as their input and output key-value classes.
	 * @param conf configuration parameters.
	 */
	public static void setup(JobConf conf) {
		conf.setJobName("epic");
		conf.setNumReduceTasks(1);
		conf.setMapperClass(CountMapper.class);
		conf.setReducerClass(CountReducer.class);
		conf.setMapOutputKeyClass(MultiFieldKey.class);
		conf.setMapOutputValueClass(BigIntegerWritable.class);
		conf.setOutputKeyClass(MultiFieldKey.class);
		conf.setOutputValueClass(BigIntegerWritable.class);

		conf.setInputFormat(RecordInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);

	}

}
