package mapred;

import java.io.IOException;
import java.util.Iterator;

import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextOutputFormat;

import common.Cipher;
import common.MultiArray;
import common.MultiFieldKey;
import common.Record;

/**
 * This is similar to {@link MapRedPlainCountOne}, but supports counting many values at once.
 * Still, it does not support disjunctive, arbitrary boolean expression, GF(2) arithmetized
 * counting.
 * 
 * @author vohuudtr
 *
 */
public class MapRedPlainCountAll {

	private static class CountMapper extends MapReduceBase implements
			Mapper<LongWritable, Record, MultiFieldKey, BigIntegerWritable> {
		private MultiArray<BigIntegerWritable> sum;
		private OutputCollector<MultiFieldKey, BigIntegerWritable> output;
		private int numFields;
		private int[] size_D;

		public void configure(JobConf conf) {
			Cipher.initialize(Count.class.getResourceAsStream(conf
					.get("paramfile")));

			numFields = conf.getInt("numFields", 0);
			size_D = new int[numFields];
			int capacity = 1;
			for (int i = 0; i < numFields; i++) {
				size_D[i] = conf.getInt("|D" + i + "|", 0);
				capacity *= size_D[i];
			}
			sum = MultiArray.createBigIntegerWritableArray(capacity);
		}

		public void map(LongWritable key, Record record,
				OutputCollector<MultiFieldKey, BigIntegerWritable> output,
				Reporter reporter) throws IOException {
			this.output = output;
			computeSumUnencrypted(record);
		}

		public void computeSumUnencrypted(Record record) {
			MultiFieldKey key = new MultiFieldKey(numFields);
			for (int j = 0; j < numFields; j++) {
				int v = Integer.parseInt(record.getField(j * 2 + 1).toString());
				key.element[j] = v;
			}
			BigIntegerWritable lastsum = sum.get(key);
			if (lastsum == null)
				sum.put(key, new BigIntegerWritable(1));
			else
				sum.put(key, lastsum.add(new BigIntegerWritable(1)));
		}

		public void writeSum() {
			BigIntegerWritable[] items = sum.getAll();
			for (int i = 0; i < items.length; i++) {
				try {
					if (items[i] != null)
						output.collect(MultiArray.getMultiFieldKey(i), items[i]);
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

		public void close() {
			writeSum();
		}
	}

	private static class CountReducer extends MapReduceBase
			implements
			Reducer<MultiFieldKey, BigIntegerWritable, MultiFieldKey, BigIntegerWritable> {
		public void reduce(MultiFieldKey key,
				Iterator<BigIntegerWritable> values,
				OutputCollector<MultiFieldKey, BigIntegerWritable> output,
				Reporter reporter) throws IOException {
			BigIntegerWritable sum = new BigIntegerWritable(0);
			while (values.hasNext()) {
				BigIntegerWritable value = values.next();
				sum = sum.add(value);
			}
			output.collect(key, sum);
		}
	}

	/**
	 * Initializes the class based on provided configuration parameters.
	 * The initialization comprises setting job's name, assigning Mapper
	 * and Reducer class as well as their input and output key-value classes.
	 * @param conf configuration parameters.
	 */
	public static void setup(JobConf conf) {
		conf.setJobName("plain");
		//		conf.setNumReduceTasks(1);
		conf.setMapperClass(CountMapper.class);
		conf.setReducerClass(CountReducer.class);
		conf.setMapOutputKeyClass(MultiFieldKey.class);
		conf.setMapOutputValueClass(BigIntegerWritable.class);
		conf.setOutputKeyClass(MultiFieldKey.class);
		conf.setOutputValueClass(BigIntegerWritable.class);

		conf.setInputFormat(RecordInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);

	}

}
