package mapred;

import java.io.DataInput;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Iterator;

import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextOutputFormat;

import common.MultiFieldKey;
import common.Record;

/**
 * This is an illustrating implementation of counting based on unencrypted fields.
 * This implementation supports single field counting and multiple conjuctive counting
 * only. Disjunctive counting (and therefore, arbitrary boolean expressions) is not
 * supported.
 * 
 * @author vohuudtr
 * 
 */
public class MapRedPlainCountOne {

	private static class CountMapper extends MapReduceBase implements
			Mapper<LongWritable, Record, MultiFieldKey, IntWritable> {
		private int sum;
		private OutputCollector<MultiFieldKey, IntWritable> output;
		private int numFields;
		private Path[] sharedFiles;
		private MultiFieldKey countedKey = null;

		public void configure(JobConf conf) {
			// read key to be counted
			try {
				sharedFiles = DistributedCache.getLocalCacheFiles(conf);
				DataInput request = Count.getInput(sharedFiles, "request");
				if (request == null)
					throw new RuntimeException("No counting value specified.");
				else {
					countedKey = new MultiFieldKey(request);
				}
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}

			numFields = conf.getInt("numFields", 0);
			sum = 0;
		}

		public void map(LongWritable key, Record record,
				OutputCollector<MultiFieldKey, IntWritable> output,
				Reporter reporter) throws IOException {
			this.output = output;
			computeSumUnencrypted(record);
		}

		public void computeSumUnencrypted(Record record) {
			MultiFieldKey key = new MultiFieldKey(numFields);
			for (int j = 0; j < numFields; j++) {
				int v = Integer.parseInt(record.getField(j * 2 + 1).toString());
				key.element[j] = v;
			}
			if (key.equals(countedKey))
				sum++;
		}

		public void writeSum() {
			try {
				output.collect(countedKey, new IntWritable(sum));
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		public void close() {
			writeSum();
		}
	}

	private static class CountReducer extends MapReduceBase
			implements
			Reducer<MultiFieldKey, IntWritable, NullWritable, IntWritable> {
		public void reduce(MultiFieldKey key,
				Iterator<IntWritable> values,
				OutputCollector<NullWritable, IntWritable> output,
				Reporter reporter) throws IOException {
			int sum = 0;
			while (values.hasNext()) {
				IntWritable value = values.next();
				sum += value.get();
			}
			output.collect(NullWritable.get(), new IntWritable(sum));
		}
	}

	/**
	 * Initializes the class based on provided configuration parameters.
	 * The initialization comprises setting job's name, assigning Mapper
	 * and Reducer class as well as their input and output key-value classes.
	 * @param conf configuration parameters.
	 */
	public static void setup(JobConf conf) {
		conf.setJobName("plain");
		conf.setNumReduceTasks(1);
		conf.setMapperClass(CountMapper.class);
		conf.setReducerClass(CountReducer.class);
		conf.setMapOutputKeyClass(MultiFieldKey.class);
		conf.setMapOutputValueClass(IntWritable.class);
		conf.setOutputKeyClass(MultiFieldKey.class);
		conf.setOutputValueClass(IntWritable.class);

		conf.setInputFormat(RecordInputFormat.class);
		conf.setOutputFormat(TextOutputFormat.class);

	}

}
