package mapred;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.InputFormat;
import org.apache.hadoop.mapred.InputSplit;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.RecordReader;
import org.apache.hadoop.mapred.RecordWriter;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.util.Progressable;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import common.Cipher;
import common.Parameters;
import common.Record;

/**
 * This is a MapReduce job used for generating a large data of set and storing in the HDFS.
 * The generated data in the HDFS will be read later by a counting job. This class was created
 * and used for experiments.
 * <p>
 * Usage:
 * <p>
 * <code>hadoop jar &lt;JARFILE&gt; %mapred.Producer -Dquantity=&lt;quantity&gt; -Dparamfile=&lt;paramfile&gt; &lt;genoutput&gt;</code>
 * <dl>
 * <dt>JARFILE</dt><dd>The JAR file containing the application.</dd>
 * <dt>quantity</dt><dd>Number of records to generate.</dd>
 * <dt>paramfile</dt><dd>Name of parameters file. Path to the parameters file must be a relative path to this class in the JAR file.
 * See {@link Parameters} for details on parameters file.</dd>
 * <dt>genoutput</dt><dd>A HDFS path to the directory containing the generated data.</dd>
 * </dl>
 * For details on what data is generated and how records are formatted, see {@link common.Producer}. 
 * @author vohuudtr
 *
 */
public class Producer extends Configured implements Tool {

	private static class Map extends MapReduceBase implements
			Mapper<NullWritable, NullWritable, NullWritable, Record> {

		public void configure(JobConf conf) {
			Cipher.initialize(Producer.class.getResourceAsStream(conf
					.get("paramfile")));
			common.Producer.getStats().initialize(Parameters.getTotalSizeD());
		}

		public void map(NullWritable key, NullWritable value,
				OutputCollector<NullWritable, Record> output, Reporter reporter)
				throws IOException {

			// write
			output.collect(NullWritable.get(), common.Producer.generateRecord());
		}
	}

	private static class AmazonProducerInputFormat implements
			InputFormat<NullWritable, NullWritable> {

		@Override
		public RecordReader<NullWritable, NullWritable> getRecordReader(
				InputSplit split, JobConf conf, Reporter reporter)
				throws IOException {
			return new AmazonProducerRecordReader((AmazonProducerSplit) split);
		}

		@Override
		public InputSplit[] getSplits(JobConf conf, int numSplits)
				throws IOException {
			int quantity = conf.getInt("quantity", 0);
			int numRecsPerSplit = (int) quantity / numSplits;

			InputSplit splits[] = new InputSplit[numSplits];

			for (int i = 0; i < numSplits; i++) {
				if (i == numSplits - 1)
					splits[i] = new AmazonProducerSplit(i, quantity - i
							* numRecsPerSplit);
				else
					splits[i] = new AmazonProducerSplit(i, numRecsPerSplit);
			}

			return splits;
		}
	}

	private static class AmazonProducerRecordReader implements
			RecordReader<NullWritable, NullWritable> {
		private long counter = 0;
		private int numRecs;

		public AmazonProducerRecordReader(AmazonProducerSplit split) {
			numRecs = split.numRecsPerSplit;
		}

		@Override
		public void close() throws IOException {
		}

		@Override
		public NullWritable createKey() {
			return NullWritable.get();
		}

		@Override
		public NullWritable createValue() {
			return NullWritable.get();
		}

		@Override
		public long getPos() throws IOException {
			return counter;
		}

		@Override
		public float getProgress() throws IOException {
			return Math.min(1.0f, (float) (counter + 1) / numRecs);
		}

		@Override
		public boolean next(NullWritable arg0, NullWritable arg1)
				throws IOException {
			if (counter < numRecs) {
				counter++;
				return true;
			}
			return false;
		}

	}

	private static class AmazonProducerSplit implements InputSplit {
		private int fileIndex;
		private int numRecsPerSplit;

		@SuppressWarnings("unused")
		public AmazonProducerSplit() {

		}

		public AmazonProducerSplit(int fileIndex, int numRecsPerSplit) {
			this.fileIndex = fileIndex;
			this.numRecsPerSplit = numRecsPerSplit;
		}

		@Override
		public void readFields(DataInput in) throws IOException {
			fileIndex = in.readInt();
			numRecsPerSplit = in.readInt();
		}

		@Override
		public void write(DataOutput out) throws IOException {
			out.writeInt(fileIndex);
			out.writeInt(numRecsPerSplit);
		}

		@Override
		public long getLength() throws IOException {
			return numRecsPerSplit * Record.blocksize;
		}

		@Override
		public String[] getLocations() throws IOException {
			return new String[] {};
		}
	}

	private static class AmazonProducerOutputFormat extends
			FileOutputFormat<NullWritable, Record> {

		@Override
		public RecordWriter<NullWritable, Record> getRecordWriter(
				FileSystem fs, JobConf conf, String name, Progressable progress)
				throws IOException {
			return new AmazonProducerRecordWriter(fs.create(
					FileOutputFormat.getTaskOutputPath(conf,
							name.replaceAll("part", "data")), progress));
		}

	}

	private static class AmazonProducerRecordWriter implements
			RecordWriter<NullWritable, Record> {
		private DataOutputStream out;

		public AmazonProducerRecordWriter(DataOutputStream out) {
			this.out = out;
		}

		public synchronized void close(Reporter arg0) throws IOException {
			out.close();
		}

		@Override
		public void write(NullWritable arg0, Record record) throws IOException {
			record.write(out);
		}

	}

	/**
	 * main function
	 */
	public int run(String[] args) throws Exception {
		JobConf conf = new JobConf(getConf(), Producer.class);
		FileSystem fs = FileSystem.get(conf);

		conf.setJobName(conf.get("name", "producer"));

		// set classes depending on encrypted or plain version
		conf.setMapperClass(Map.class);
		conf.setNumReduceTasks(0);
		conf.setMapOutputKeyClass(NullWritable.class);
		conf.setMapOutputValueClass(Record.class);
		conf.setInputFormat(AmazonProducerInputFormat.class);
		conf.setOutputFormat(AmazonProducerOutputFormat.class);

		// delete old output before running job
		fs.delete(new Path(args[0]), true);

		FileOutputFormat.setOutputPath(conf, new Path(args[0]));

		JobClient.runJob(conf);

		return 0;
	}

	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new Producer(), args);
		System.exit(res);
	}
}
